"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
var DTO_1 = __importDefault(require("./util/DTO"));
var Logger_1 = __importDefault(require("./util/Logger"));
/**
 * Classe de acesso aos dados.
 *
 * O resultado de todas as mensagens de tópicos são persistidos OFFLINE, com isso, evita tráfego desnecessário.
 *
 * Implementa um mecanismo de sincronização com o servidor, informando ao mesmo sobre a ultima versão recebida de cada tópico
 */
var Client = /** @class */ (function () {
    function Client(host, storage) {
        this.reconnect = true;
        /**
         * Deixa salvo as inscrições em tópicos, em problema de conexão, solicita novamente a subscriçao
         */
        this.subscriptions = {};
        /**
         * Permite receber resposta para uma execução
         */
        this.promises = {};
        this.timeouts = [];
        this.host = host;
        this.storage = storage;
        this.LOGGER = Logger_1.default.get("Client");
    }
    /**
     * Quando invocado, não será mais criado novas conexões
     */
    Client.prototype.close = function () {
        this.LOGGER.trace('Fechando conexão');
        this.reconnect = false;
        //@TODO: Limpar todas as referencias para callbacks
        // this.subscriptions = {};
        // Desconecta-se do servidor
        if (this.ws) {
            this.ws.close();
            this.ws = undefined;
        }
        // remove todos os timeouts
        this.timeouts.forEach(function (id) {
            clearTimeout(id);
        });
        this.timeouts = [];
    };
    /**
     * Inicializa a conexão do DataLayer com o Backend
     */
    Client.prototype.connect = function (callback) {
        var _this = this;
        if (this.ws && (this.ws.readyState === WebSocket.CONNECTING || this.ws.readyState === WebSocket.OPEN)) {
            return;
        }
        this.reconnect = true;
        this.ws = new WebSocket(this.host);
        this.LOGGER.trace('Iniciando conexão. { host=', this.host, ' }');
        this.ws.onopen = function (event) {
            if (_this.ws) {
                _this.LOGGER.trace('Conexão iniciada com sucesso');
                // connection opened
                _this.syncTopics();
                if (callback) {
                    callback();
                }
            }
        };
        this.ws.onmessage = function (event) {
            var data;
            _this.LOGGER.trace('Mensagem recebida. { data=', event.data, ' }');
            try {
                data = JSON.parse(event.data);
            }
            catch (e) {
                _this.LOGGER.error('Erro ao processar mensagem', e.data);
                return;
            }
            if (data.id) {
                // Resposta de uma ação
                data = data;
                // Está respondendo a uma solicitação
                if (!_this.promises[data.id]) {
                    // Promise já foi resolvido
                    return;
                }
                // Resolve o promise
                if (data.error) {
                    _this.promises[data.id].reject(data.error);
                }
                else {
                    _this.promises[data.id].accept(data.data);
                }
                // Remove a referencia para o promise
                delete _this.promises[data.id];
            }
            else if (data.topic) {
                // Mensagem de um tópico
                var message_1 = data;
                var subscription_1 = _this.subscriptions[message_1.topic];
                if (!subscription_1) {
                    // Atualiza o servidor com as informações sobre os tópicos que este deseja ouvir
                    _this.syncTopics();
                    return;
                }
                if (subscription_1.seq >= message_1.seq) {
                    // A mensagem está atrasada, o servidor já enviou outra mensagem mais recente neste tópico antes
                    _this.syncTopic(message_1.topic);
                    return;
                }
                var update = false;
                // ==================================================================
                // Algoritmo de Compactação dos Dados
                // Garantir o transporte apenas do essencial
                // ==================================================================
                //  1 - DTO (arrays por id)
                //  2 - Diff do flatten (geração do delta)
                //  3 - Compressão 
                // ==================================================================
                if (message_1.delta) {
                    _this.LOGGER.trace('Delta de mensagem recebida');
                    if (message_1.deltaSeq === subscription_1.seq) {
                        // É o diff correto
                        update = true;
                        subscription_1.seq = message_1.seq;
                        subscription_1.dto = subscription_1.dto.patchCompressed(message_1.delta);
                    }
                    else {
                        // Delta inválido, pode ser mensagem atrasada
                        _this.syncTopic(message_1.topic);
                    }
                }
                else {
                    _this.LOGGER.trace('Mensagem completa recebida');
                    // Mensagem completa
                    update = true;
                    subscription_1.seq = message_1.seq;
                    subscription_1.dto = DTO_1.default.decompress(message_1.data);
                }
                if (update) {
                    subscription_1.data = subscription_1.dto.unflatten();
                    ;
                    // Entrega mensagem aos interessados
                    subscription_1.callbacks.forEach(function (callback) {
                        // callback(data);
                        callback(subscription_1.data);
                    });
                    // Sincroniza com o servidor sobre a ultima mensagem recebida no tópico
                    _this.syncTopic(message_1.topic);
                    // Salva no Storage os dados do tópico
                    // Só persiste quando recebe novas mensagens
                    _this.setTimeout(function () {
                        _this.storage.set("topic_" + message_1.topic, {
                            seq: subscription_1.seq,
                            compressed: subscription_1.dto.compress(),
                        });
                    }, 0);
                }
            }
        };
        this.ws.onerror = function (event) {
            // an error occurred
            _this.LOGGER.error('Erro na conexão com o servidor', event);
        };
        this.ws.onclose = function (event) {
            _this.LOGGER.trace('Conexão WebSocket finalizada. { reason="', event.reason, '", code=', event.code, ' }');
            // connection closed
            if (_this.reconnect) {
                _this.LOGGER.trace('Reiniciando a conexão com o WebSocket');
                // Faz nova conexão
                _this.setTimeout(_this.connect.bind(_this), 10);
            }
        };
    };
    /**
     * Aguarda uma mensagem específica do servidor (exemplo. wsLoginSucces)
     *
     * @param topic
     */
    Client.prototype.subscribe = function (topic, callback) {
        var _this = this;
        if (this.isClosed()) {
            return function () { };
        }
        this.LOGGER.trace('Executando a inscrição em um tópico: { topic=', topic, ' }');
        if (!this.subscriptions[topic]) {
            this.subscriptions[topic] = {
                callbacks: [],
                data: undefined,
                seq: 0
            };
        }
        this.subscriptions[topic].callbacks.push(callback);
        // Se já possuir dados (OFFLINE), já entrega ao solicitante 
        if (this.subscriptions[topic].data) {
            this.LOGGER.trace('Possui dados offline já salvo para o tópico: { topic=', topic, ' }');
            callback(this.subscriptions[topic].data);
            // Sincroniza os tópicos de interesse
            this.syncTopics();
        }
        else {
            // Verifica se no Storage possui cache da mensagem
            this.storage.get("topic_" + topic)
                .then(function (data) {
                var idx = _this.subscriptions[topic].callbacks.indexOf(callback);
                // Se houver dados no storage e ainda não recebeu a mensagem do server, e ainda o callback é valido
                // Só invoca o callback se esse subscription não tiver sido removido
                if (data && !_this.subscriptions[topic].data && idx >= 0) {
                    // Preenche os dados do tópico
                    var storageTopic = data;
                    var subscription = _this.subscriptions[topic];
                    subscription.seq = storageTopic.seq;
                    subscription.dto = DTO_1.default.decompress(storageTopic.compressed);
                    subscription.data = subscription.dto.unflatten();
                    _this.LOGGER.trace('Possui dados offline no storage do tópico: { topic=', topic, ' }');
                    callback(_this.subscriptions[topic].data);
                }
                // Sincroniza os tópicos de interesse
                _this.syncTopics();
            })
                .catch(function (err) {
                // Sincroniza os tópicos de interesse
                _this.syncTopics();
            });
        }
        var canceled = false;
        // Cancelable
        return function () {
            if (canceled) {
                // Evita processamento desnecessário
                return;
            }
            _this.LOGGER.trace('Inscrição no tópico foi cancelado: { topic=', topic, ' }');
            canceled = true;
            // Remove o listener
            var idx = _this.subscriptions[topic].callbacks.indexOf(callback);
            if (idx >= 0) {
                _this.subscriptions[topic].callbacks.splice(idx, 1);
            }
            // Atualiza os topicos de interesse
            _this.syncTopics();
        };
    };
    /**
     * Permite executar alguma ação no server (Ex.: do('login', nomeDeUsuario))
     *
     * @param action
     */
    Client.prototype.exec = function (action, data) {
        var _this = this;
        if (this.isClosed()) {
            return Promise.reject('Client is closed');
        }
        this.LOGGER.trace('Executando uma ação { action=', action, ', data=', data, ' }');
        var requestId = "" + Client.REQUEST_SEQUENCE.id++;
        var promise = new Promise(function (accept, reject) {
            if (_this.ws) {
                _this.promises[requestId] = {
                    accept: accept,
                    reject: reject
                };
                _this.ws.send(JSON.stringify({
                    id: requestId,
                    action: action,
                    data: data
                }));
            }
            else {
                reject('Não está conectado à API');
            }
        });
        // Após 30 segundos, se não hover resposta, Timeout
        this.setTimeout(function () {
            if (_this.promises[requestId]) {
                _this.promises[requestId].reject('A resposta da requisição excedeu 30 segundos.');
                delete _this.promises[requestId];
            }
        }, 30000);
        return promise;
    };
    /**
     * Informa ao servidor quais os tópicos que essa conexão está ouvindo
     * Evita que o servidor envie todas as mensagens para todos
     */
    Client.prototype.syncTopics = function () {
        var _this = this;
        if (this.isClosed()) {
            return;
        }
        this.LOGGER.trace('Sincronizando todos os tópicos');
        var data = Object.keys(this.subscriptions)
            .map(function (topic) {
            // Se não existe callback, não está subscrito
            if (_this.subscriptions[topic].callbacks.length < 1) {
                return undefined;
            }
            return {
                topic: topic,
                seq: _this.subscriptions[topic].seq
            };
        })
            .filter(function (item) { return item !== undefined; });
        this.exec('syncTopics', data)
            .then(function () { })
            .catch(function () { });
    };
    /**
     * Informa ao servidor quais os tópicos que essa conexão está ouvindo
     * Evita que o servidor envie todas as mensagens para todos
     */
    Client.prototype.syncTopic = function (topic) {
        if (this.isClosed()) {
            return;
        }
        this.LOGGER.trace('Sincronizando um tópico específico: { topic=', topic, ' }');
        var data = {
            topic: topic,
            seq: this.subscriptions[topic].seq
        };
        this.exec('syncTopic', data)
            .then(function () { })
            .catch(function () { });
    };
    /**
     * Timeouts controlados
     *
     * @param handler
     * @param timeout
     */
    Client.prototype.setTimeout = function (handler, timeout) {
        var _this = this;
        if (this.isClosed()) {
            return -1;
        }
        var id = setTimeout(function () {
            var idx = _this.timeouts.indexOf(id);
            if (idx >= 0) {
                _this.timeouts.splice(idx, 1);
            }
            handler();
        }, timeout);
        this.timeouts.push(id);
        return id;
    };
    Client.prototype.isClosed = function () {
        return (!this.ws && !this.reconnect);
    };
    /**
     * Identificador das requisições feitas ao servidor
     */
    Client.REQUEST_SEQUENCE = { id: Date.now() };
    return Client;
}());
exports.default = Client;
//# sourceMappingURL=Client.js.map