"use strict";
var __assign = (this && this.__assign) || function () {
    __assign = Object.assign || function(t) {
        for (var s, i = 1, n = arguments.length; i < n; i++) {
            s = arguments[i];
            for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p))
                t[p] = s[p];
        }
        return t;
    };
    return __assign.apply(this, arguments);
};
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (Object.hasOwnProperty.call(mod, k)) result[k] = mod[k];
    result["default"] = mod;
    return result;
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
var WebSocket = __importStar(require("ws"));
var Topic_1 = __importDefault(require("./Topic"));
var Actions_1 = __importDefault(require("./Actions"));
var MidlewareManager_1 = __importDefault(require("./util/MidlewareManager"));
var stoppable = require('stoppable');
// Import default actions
require("./actions/syncTopicAction");
require("./actions/syncTopicsAction");
/**
 * Servidor do Websocket-pubsub, faz gerenciamento de conexões e todo o tratamento necessário para o funcionamento da biblioteca
 */
var Server = /** @class */ (function () {
    function Server(server, options, callback) {
        this.middlewares = new MidlewareManager_1.default();
        this.onConnection = this.onConnection.bind(this);
        this.server = stoppable(server);
        //initialize the WebSocket server instance
        this.webSocketServer = new WebSocket.Server(__assign({}, (options || {}), { server: server }), callback);
        this.webSocketServer.on('connection', this.onConnection);
    }
    /**
     * Finaliza o webSocketServer e o http.Server usados
     *
     * @param cb
     */
    Server.prototype.close = function (cb) {
        this.server.stop(cb);
    };
    /**
     * Permite adicionar um MidleWare na conexão.
     *
     * Util para aplicar autenticação e outros controles
     *
     * @param middleware
     */
    Server.prototype.use = function (middleware, event) {
        this.middlewares.add(middleware, event);
    };
    /**
     * Permite remover um Midleware adicionado anteriormente
     *
     * @param event
     * @param middleware
     */
    Server.prototype.removeMidleware = function (middleware, event) {
        this.middlewares.remove(middleware, event);
    };
    Server.prototype.onConnection = function (ws, request) {
        var _this = this;
        this.middlewares.exec('connection', { ws: ws, request: request, event: 'connection' })
            .then(function () {
            var context = {
                ws: ws,
                request: request
            };
            ws.on('close', _this.createHandleOnClose(context));
            ws.on('upgrade', _this.createHandleOnUpgrade(context));
            ws.on('message', _this.createHandleOnMessage(context));
        })
            .catch(function (err) {
            // Erro na conexao, remove todos os listeners            
            Topic_1.default.ALL.forEach(function (topic) {
                topic.unsubscribe(ws);
            });
        });
    };
    Server.prototype.createHandleOnUpgrade = function (context) {
        var _this = this;
        return function (uws, urequest) {
            context.ws = uws;
            context.request = urequest;
            _this.middlewares.exec('upgrade', {
                event: 'upgrade',
                ws: context.ws,
                request: context.request
            });
        };
    };
    Server.prototype.createHandleOnClose = function (context) {
        var _this = this;
        return function (code, reason) {
            // Ao desconectar, remove o socket de todos os tópicos
            Topic_1.default.ALL.forEach(function (topic) {
                topic.unsubscribe(context.ws);
            });
            _this.middlewares.exec('close', {
                event: 'close',
                ws: context.ws,
                request: context.request,
                code: code,
                reason: reason
            });
        };
    };
    Server.prototype.createHandleOnMessage = function (context) {
        var _this = this;
        return function (data) {
            var midContext = {
                event: 'message',
                ws: context.ws,
                request: context.request,
                data: data
            };
            _this.middlewares.exec('message', midContext).then(function () {
                var request;
                // Algum midleware pode modificar os dados de entrada.
                // Desde que mantenha a estrutura, é permitido
                if (typeof midContext.data === 'string') {
                    var msg = midContext.data;
                    try {
                        request = JSON.parse(msg);
                    }
                    catch (e) {
                        console.warn('Erro ao processar mensagem', msg);
                        return;
                    }
                }
                else {
                    request = midContext.data;
                }
                if (!request.action) {
                    // A única forma de conexão com o server, a partir daqui, é via actions
                    return;
                }
                _this.execAction(context, request);
            });
        };
    };
    /**
     * Faz a execução dos Midelewares e Tratamento de um action
     */
    Server.prototype.execAction = function (context, request) {
        Actions_1.default.exec(context, request)
            .then(function (data) {
            // Envia a resposta ao solicitante
            if (context.ws.readyState === context.ws.OPEN) {
                context.ws.send(JSON.stringify({
                    id: request.id,
                    data: data
                }));
            }
        })
            .catch(function (cause) {
            // Envia o erro ao solicitante
            if (context.ws.readyState === context.ws.OPEN) {
                context.ws.send(JSON.stringify({
                    id: request.id,
                    error: "" + (cause || 'Ocorreu um erro inesperado.'),
                    stack: cause ? cause.stack : undefined
                }));
            }
        });
    };
    return Server;
}());
exports.default = Server;
//# sourceMappingURL=Server.js.map