import * as WebSocket from 'ws';
import DTO from './util/DTO';
import Repository from './repository/Repository';
/**
 * Mensagens provenientes do servidor
 */
export interface TopicResponse {
    /**
     * Tópico que disparou a mensagem
     */
    topic: string;
    /**
     * Sequencial da mensagem atual
     */
    seq: number;
    /**
    * Corpo da mensagem, quando enviado todo o conteúdo para o cliente
    *
    * DTO::compress()
    */
    data?: any;
    /**
    * Diff, quando o cliente já possui dados atualizados (3 ultimas atualizações), envia apenas um patch, que será aplicado pelo cliente
    *
    * Delta::compress()
    */
    delta?: any;
    /**
     * ID da mensagem que gerou o diff
     */
    deltaSeq?: number;
}
/**
 * Informação de um cliente deste tópico
 */
export interface TopicSubscriber {
    /**
     * Conexão com o Subscrito
     */
    ws: WebSocket;
    /**
     * Última mensagem enviada ao cliente
     */
    lastSeq: number;
    /**
     * Timestamp da ultima mensagem enviada
     */
    lastTs: number;
    /**
     * Última mensagem confirmada pelo cliente
     */
    lastReceivedSeq: number;
}
/**
 * Representação dos dados de um tópico
 */
export interface TopicData {
    /**
     * O número sequencial desta mensagem
     */
    seq: number;
    /**
     * Mensagem original deste sequencial
     *
     * DTO
     */
    dto?: DTO;
    /**
     * Mensagem original deste sequencial comprimido, é persistido na base de dados e também é recuperado durante o carregamento
     *
     * DTO::compress()
     */
    dataCompressed: string;
}
/**
 * Representação dos deltas de um tópico
 */
export interface TopicDelta extends TopicData {
    /**
     * Diferença entre o estado atual dos dados e o estado da mensagem deste delta
     *
     * Delta::compress()
     */
    diffCompressed: string;
}
/**
 * Representação do estado atual deste tópico
 */
export interface TopicState extends TopicData {
    /**
     * O nome deste tópico
     */
    name: string;
    /**
     * Calcula e salva o Diff das ultimas <NUM_DELTAS_HISTORY> mensagens no tópico.
     *
     * Usado para envio rápido a usuários com atraso de rede temporário
     */
    deltas: Array<TopicDelta>;
}
/**
 * Tópico genérico
 *
 * @TODO: Permitir adicionar Midlewares nos tópicos
 */
export default class Topic {
    /**
     * Mantém referencia para todos os tópicos criados
     */
    static ALL: Array<Topic>;
    /**
     * Os nomes de tópicos registrados (que tem publishers)
     */
    static REGISTERED_NAMES: Array<string>;
    /**
     * Permite persistir e recuperar o estado das mensagens do tópico da base de dados
     */
    static repository?: Repository<any>;
    /**
     * Define o mecanismo de backup e restauração do tópico
     */
    static setStorage: (storage: Repository<any>) => void;
    /**
     * Obtém a instancia de um topico por nome. Se não exisitr, faz a instanciação do mesmo
     */
    static find: (name: string) => Topic | undefined;
    /**
     * Faz o envio em um tópico por nome
     *
     * Se o tópico não existir em memoria, é instanciado um
     */
    static send: (name: string, data: any) => void;
    /**
     * Sequencial das mensagens deste tópico
     */
    private SEQ;
    /**
     * Faz cache de mensagens por sequencial e deltaSequencial, evita invocar JSON.stringify todo o tempo
     */
    private stringifyCached;
    /**
     * O nome do tópico, usado na mensageria
     */
    private name;
    /**
     * A cada <RESEND_INTERVAL> segundos envia as mensagens pendentes (clientes que não confirmaram o recebimento)
     */
    private interval;
    /**
     * Proxima mensagen a ser enviada.
     *
     * Pode ocorrer de ser solicitado o envio de novos dados antes de finalizar o envio atual. Registra em memoria os novos dados para reenviar
     * quando o envio atual finalizar.
     */
    private next?;
    /**
     * O estado atual do Tópico, contendo a mesagem atual e o diff das ultmas <NUM_DELTAS_HISTORY> mensagens
     */
    private state;
    /**
    * Indica que está efetuando o carregamento dos dados historicos
    */
    private isLoading;
    /**
     * Indica que está enviando dados neste momento
     */
    private isSending;
    /**
     * Os subscritos neste tópico
     */
    private subscribers;
    constructor(name: string);
    getName(): string;
    /**
     * Permite enviar dados para o tópico
     *
     * O dado só será enviado se houver diferença entre o novo dado e o registro anterior.
     *
     * @param newData
     */
    send(newData: any): void;
    /**
     * Registra um cliente neste tópico
     *
     * @param ws
     * @param lastReceivedSeq ID da ultma mensagem recebida pelo cliente
     */
    subscribe(ws: WebSocket, lastReceivedSeq: number): void;
    /**
     * Remove um subscripto deste tópico
     *
     * @param ws
     */
    unsubscribe(ws: WebSocket): void;
    /**
     * Envia para todos os que não receberam ou não confirmaram o recebimento da ultima mensagem
     */
    private sendToAllSubscribers;
    /**
     * Envia a mensage para o cliente, se necessário.
     *
     * Valida se o cliente já possui a versão mais recente e envia apenas a informação necessária, como o diff ou conteúdo completo
     *
     * @param ws
     */
    private sendToSubscriberIfNeed;
    /**
     * Inicializa a tarefa de reenvio para clientes que não responderam a tempo
     */
    private initializeResendJob;
    /**
     * Obtém da base de dados o estado inicial do Topic
     */
    private loadFromStorage;
    /**
     * Persiste o estado atual do tópico na base de dados
     */
    private persist;
}
