"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
var DTO_1 = __importDefault(require("./util/DTO"));
/**
 * Tenta enviar para os clientes que não responderam neste intervalo
 */
var RESEND_TIMEOUT = 3000;
/**
 * Quantidade de delta mantido no historico
 */
var NUM_DELTAS_HISTORY = 5;
/**
 * Tópico genérico
 *
 * @TODO: Permitir adicionar Midlewares nos tópicos
 */
var Topic = /** @class */ (function () {
    function Topic(name) {
        /**
         * Sequencial das mensagens deste tópico
         */
        this.SEQ = 0;
        /**
         * O estado atual do Tópico, contendo a mesagem atual e o diff das ultmas <NUM_DELTAS_HISTORY> mensagens
         */
        this.state = {
            name: this.getName(),
            seq: 0,
            dto: undefined,
            dataCompressed: '',
            deltas: []
        };
        /**
        * Indica que está efetuando o carregamento dos dados historicos
        */
        this.isLoading = false;
        /**
         * Indica que está enviando dados neste momento
         */
        this.isSending = false;
        /**
         * Os subscritos neste tópico
         */
        this.subscribers = [];
        // Verifica se o nome do topico é valido
        if (!Topic.REGISTERED_NAMES.find(function (validName) { return name.split('#')[0] === validName; })) {
            throw new Error('Não existe um Publisher registrado para esse tópico');
        }
        this.name = name;
        // Já inicia o carregamento do estado atual do tópico
        this.loadFromStorage();
        // Registra na lista de tópicos o item atual
        Topic.ALL.push(this);
    }
    Topic.prototype.getName = function () {
        return this.name;
    };
    /**
     * Permite enviar dados para o tópico
     *
     * O dado só será enviado se houver diferença entre o novo dado e o registro anterior.
     *
     * @param newData
     */
    Topic.prototype.send = function (newData) {
        var _this = this;
        if (this.isLoading || this.isSending) {
            // Adiciona na fila de envio
            this.next = newData;
            return;
        }
        this.isSending = true;
        // Garante que não possui nova mensagen na fila
        this.next = undefined;
        var flattenData = new DTO_1.default(newData);
        // Novo estado do tópico
        var newState = {
            name: this.getName(),
            seq: -1,
            dto: flattenData,
            dataCompressed: flattenData.compress(),
            deltas: []
        };
        if (this.state.dto) {
            // Já possui uma mensagem atual
            // Verifica se tem alterações
            // ==================================================================
            // Algoritmo de Compactação dos Dados
            // Garantir o transporte apenas do essencial
            // ==================================================================
            //  1 - DTO (arrays por id)
            //  2 - Diff do flatten (geração do delta)
            //  3 - Compressão
            // ==================================================================
            var delta = this.state.dto.diff(newState.dto);
            if (!delta.hasDiff()) {
                // Não possui alterações, ignora o versionamento e envio
                this.isSending = false;
                return;
            }
            // Garante imutabilidade do registro
            newState.deltas = this.state.deltas.slice(0);
            // Ordena os deltas, os mais recentes no inicio (Os mais antigos vão ser removidos)
            newState.deltas.sort(function (a, b) { return b.seq - a.seq; });
            // Limite em apenas <NUM_DELTAS_HISTORY - 1> deltas no estado
            newState.deltas.splice(NUM_DELTAS_HISTORY - 1);
            // Calcula os novos deltas
            newState.deltas.forEach(function (delta) {
                delta.diffCompressed = delta.dto.diff(newState.dto).compress();
            });
            // Insere o novo diff no inicio
            newState.deltas.push({
                seq: this.state.seq,
                dto: this.state.dto,
                dataCompressed: this.state.dataCompressed,
                diffCompressed: delta.compress()
            });
        }
        // Possui atualização ou é primeiro registro
        // Incrementa o sequencial da ultima mensagem enviada no tópico
        newState.seq = ++this.SEQ;
        // Persiste o novo estado
        this.persist(newState)
            .then(function () {
            // Finalmente, entrega aos interessados, atualiza o estado do tópico e verifica se tem novas atualizações
            _this.isSending = false;
            // Sobrescreve o estado anterior
            _this.state = newState;
            _this.sendToAllSubscribers();
            if (_this.next) {
                // Possui novos dados para enviar
                _this.send(_this.next);
            }
        })
            .catch(function (err) {
            _this.isSending = false;
            if (_this.next) {
                // Possui novos dados para enviar
                _this.send(_this.next);
            }
        });
    };
    /**
     * Registra um cliente neste tópico
     *
     * @param ws
     * @param lastReceivedSeq ID da ultma mensagem recebida pelo cliente
     */
    Topic.prototype.subscribe = function (ws, lastReceivedSeq) {
        var subscriber = this.subscribers.find(function (subscriber) { return subscriber.ws === ws; });
        if (subscriber) {
            subscriber.lastReceivedSeq = Math.max(subscriber.lastReceivedSeq, lastReceivedSeq);
        }
        else {
            subscriber = {
                ws: ws,
                lastSeq: 0,
                lastTs: Date.now(),
                lastReceivedSeq: lastReceivedSeq
            };
            this.subscribers.push(subscriber);
        }
        // Verifica e envia para o subscrito
        this.sendToSubscriberIfNeed(subscriber);
        // Inicializa a tarefa de reenvio
        this.initializeResendJob();
    };
    /**
     * Remove um subscripto deste tópico
     *
     * @param ws
     */
    Topic.prototype.unsubscribe = function (ws) {
        var subscriber = this.subscribers.find(function (subscriber) { return subscriber.ws === ws; });
        if (subscriber) {
            var idx = this.subscribers.indexOf(subscriber);
            if (idx >= 0) {
                this.subscribers.splice(idx, 1);
            }
        }
    };
    /**
     * Envia para todos os que não receberam ou não confirmaram o recebimento da ultima mensagem
     */
    Topic.prototype.sendToAllSubscribers = function () {
        var _this = this;
        // Limpa cache de JSON.stringify
        this.stringifyCached = {};
        this.subscribers.forEach(function (subscriber) {
            _this.sendToSubscriberIfNeed(subscriber);
        });
    };
    /**
     * Envia a mensage para o cliente, se necessário.
     *
     * Valida se o cliente já possui a versão mais recente e envia apenas a informação necessária, como o diff ou conteúdo completo
     *
     * @param ws
     */
    Topic.prototype.sendToSubscriberIfNeed = function (subscriber) {
        var message = {
            topic: this.getName(),
            seq: this.state.seq
        };
        // Se a ultma mensagem confirmada pelo subscrito é anterior a mensagem atual, e já expirou o tempo de resposta
        if (subscriber.lastReceivedSeq < this.state.seq && subscriber.lastTs < Date.now() - RESEND_TIMEOUT) {
            // Se houver algum delta para a mensagem que o subscrito possui, envia esse delta
            var delta = this.state.deltas.find(function (delta) { return delta.seq === subscriber.lastReceivedSeq; });
            if (delta) {
                message.deltaSeq = delta.seq;
                message.delta = delta.diffCompressed;
            }
            else {
                // Não possui delta, envia a mensagem completa
                message.data = this.state.dataCompressed;
            }
        }
        // Se a ultima versão enviada para o subscrito é anterior a mensagem atual
        else if (subscriber.lastSeq < this.state.seq) {
            // Se houver algum delta para a mensagem que o subscrito possui, envia esse delta
            var delta = this.state.deltas.find(function (delta) { return delta.seq === subscriber.lastSeq; });
            if (delta) {
                message.deltaSeq = delta.seq;
                message.delta = delta.diffCompressed;
            }
            else {
                // Não possui delta, envia a mensagem completa
                message.data = this.state.dataCompressed;
            }
        }
        // Possui dados para enviar?
        if (message.data || message.delta) {
            subscriber.lastSeq = this.state.seq;
            // Evita invocar JSON.stringify a todo o custo
            var cacheId = message.seq + "_" + (message.deltaSeq || '');
            if (!this.stringifyCached[cacheId]) {
                this.stringifyCached[cacheId] = JSON.stringify(message);
            }
            subscriber.ws.send(this.stringifyCached[cacheId]);
        }
    };
    /**
     * Inicializa a tarefa de reenvio para clientes que não responderam a tempo
     */
    Topic.prototype.initializeResendJob = function () {
        var _this = this;
        clearInterval(this.interval);
        // A cada segundo envia novamente mensagem para os que não receberam ou não confirmaram o recebimento
        this.interval = setInterval(function () {
            if (_this.subscribers.length > 0) {
                _this.sendToAllSubscribers();
            }
            else {
                // Não possui subscribers, remover a tarefa
                clearInterval(_this.interval);
            }
        }, RESEND_TIMEOUT);
    };
    /**
     * Obtém da base de dados o estado inicial do Topic
     */
    Topic.prototype.loadFromStorage = function () {
        var _this = this;
        if (this.isLoading) {
            return;
        }
        this.isLoading = true;
        // Faz tentativas ilimitadas de loading, o Topico só fica disponível para envio quando o carregamento for efetuado
        var retries = 1;
        var tryToLoad = function () {
            if (Topic.repository) {
                Topic.repository.findOne({ name: _this.getName() }, {})
                    .then(function (row) {
                    // Marca como carregamento concluído
                    _this.isLoading = false;
                    if (row) {
                        var state = row;
                        // Extrai os dados comprimidos da base
                        _this.SEQ = state.seq;
                        _this.state.seq = state.seq;
                        _this.state.dto = DTO_1.default.decompress(state.dataCompressed);
                        _this.state.dataCompressed = state.dataCompressed;
                        _this.state.deltas = state.deltas.map(function (delta) {
                            return {
                                seq: delta.seq,
                                dataFlatten: DTO_1.default.decompress(delta.dataCompressed),
                                dataCompressed: delta.dataCompressed,
                                diffCompressed: delta.diffCompressed
                            };
                        });
                    }
                    // Verifica se já possui novas mensagens a enviar
                    if (_this.next) {
                        // Possui novos dados para enviar
                        _this.send(_this.next);
                    }
                })
                    .catch(function (err) {
                    console.error("Erro durante o carregamento do t\u00F3pico " + _this.getName() + ", " + retries + "a tentativa", err);
                    setTimeout(tryToLoad, 10);
                });
            }
            else {
                // Marca como carregamento concluído
                _this.isLoading = false;
            }
        };
        // Inicia a tentativa de carregamento
        tryToLoad();
    };
    /**
     * Persiste o estado atual do tópico na base de dados
     */
    Topic.prototype.persist = function (newState) {
        var _this = this;
        return new Promise(function (accept, reject) {
            // Faz 5 tentativas de persistir a informação, após isso, retorna erro
            var retries = 1;
            var tryToPersist = function () {
                if (Topic.repository) {
                    var repository_1 = Topic.repository;
                    // Salva apenas dados comprimidos na base
                    var dados_1 = {
                        name: _this.getName(),
                        seq: newState.seq,
                        dataCompressed: newState.dataCompressed,
                        deltas: newState.deltas.map(function (delta) {
                            return {
                                seq: delta.seq,
                                dataCompressed: delta.dataCompressed,
                                diffCompressed: delta.diffCompressed
                            };
                        }),
                    };
                    repository_1.count({ name: _this.getName() }, {})
                        .then(function (count) {
                        if (count === 0) {
                            return repository_1.insert({ name: _this.getName() }, dados_1);
                        }
                        else {
                            return repository_1.update({ name: _this.getName() }, dados_1, {});
                        }
                    })
                        .then(accept)
                        .catch(function (err) {
                        console.error("Erro ao persistir t\u00F3pico " + _this.getName() + ", " + retries + "a tentativa", err);
                        if (retries++ < 5) {
                            setTimeout(tryToPersist, 10);
                        }
                        else {
                            reject(err);
                        }
                    });
                }
                else {
                    accept();
                }
            };
            // Tenta persistir o tópico
            tryToPersist();
        });
    };
    /**
     * Mantém referencia para todos os tópicos criados
     */
    Topic.ALL = [];
    /**
     * Os nomes de tópicos registrados (que tem publishers)
     */
    Topic.REGISTERED_NAMES = [];
    /**
     * Define o mecanismo de backup e restauração do tópico
     */
    Topic.setStorage = function (storage) {
        Topic.repository = storage;
    };
    /**
     * Obtém a instancia de um topico por nome. Se não exisitr, faz a instanciação do mesmo
     */
    Topic.find = function (name) {
        var topic = Topic.ALL.find(function (topic) { return name === topic.getName(); });
        if (!topic) {
            // Tópico não existe, verifica se o nome do topico é valido
            if (Topic.REGISTERED_NAMES.find(function (validName) { return name.split('#')[0] === validName; })) {
                topic = new Topic(name);
            }
        }
        return topic;
    };
    /**
     * Faz o envio em um tópico por nome
     *
     * Se o tópico não existir em memoria, é instanciado um
     */
    Topic.send = function (name, data) {
        var topic = Topic.find(name);
        if (topic) {
            topic.send(data);
        }
    };
    return Topic;
}());
exports.default = Topic;
//# sourceMappingURL=Topic.js.map