"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
/**
 * Implementação abstrata com cache em memória para agilizar o acesso aos dados do storage.
 *
 * Os dados ficam em cache por 10 segundos apenas
 */
var ClientStorageAbstract = /** @class */ (function () {
    function ClientStorageAbstract(namespace) {
        /**
        * Cache dos dados, serializado para garantir a imutalidade
        */
        this.cache = {};
        this.promises = {};
        this.timeouts = {};
        this.namespace = namespace;
    }
    /**
     * Atualiza o prazo de vida de um item do cache
     *
     * @param key
     */
    ClientStorageAbstract.prototype.touch = function (key) {
        var _this = this;
        clearTimeout(this.timeouts[key]);
        this.timeouts[key] = setTimeout(function () {
            // Após 10 segundos sem uso, remove o item do cache
            delete _this.cache[key];
        }, 10000);
    };
    ClientStorageAbstract.prototype.get = function (key) {
        var _this = this;
        key = this.toNamespaceKey(key);
        var promise = this.promises[key];
        if (!promise) {
            promise = new Promise(function (accept, reject) {
                var resolve = function () {
                    delete _this.promises[key];
                    var data;
                    try {
                        data = JSON.parse(_this.cache[key]);
                    }
                    catch (e) {
                        reject(e);
                        return;
                    }
                    accept(data);
                };
                // busca rápida, em memória
                if (_this.cache.hasOwnProperty(key)) {
                    resolve();
                    return;
                }
                // Obtém da implementação
                _this.getItem(key)
                    .then(function (value) {
                    if (value) {
                        _this.touch(key);
                        _this.cache[key] = value;
                        resolve();
                    }
                    else {
                        delete _this.cache[key];
                        accept(null);
                    }
                })
                    .catch(function (cause) {
                    delete _this.promises[key];
                    reject(cause);
                });
            });
            this.promises[key] = promise;
        }
        return promise;
    };
    ClientStorageAbstract.prototype.set = function (key, data) {
        var _this = this;
        key = this.toNamespaceKey(key);
        return new Promise(function (accept, reject) {
            var value;
            try {
                value = JSON.stringify(data);
            }
            catch (e) {
                reject(e);
                return;
            }
            _this.setItem(key, value)
                .then(function () {
                _this.cache[key] = value;
                _this.touch(key);
                accept();
            })
                .catch(reject);
        });
    };
    ClientStorageAbstract.prototype.toNamespaceKey = function (key) {
        return "@" + this.namespace + "_" + key;
    };
    ClientStorageAbstract.prototype.remove = function (key) {
        var _this = this;
        key = this.toNamespaceKey(key);
        return new Promise(function (accept, reject) {
            _this.removeItem(key)
                .then(function (err) {
                delete _this.cache[key];
                clearTimeout(_this.timeouts[key]);
                accept();
            })
                .catch(reject);
        });
    };
    return ClientStorageAbstract;
}());
exports.default = ClientStorageAbstract;
//# sourceMappingURL=ClientStorageAbstract.js.map