"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var Compress_1 = require("./Compress");
/**
 * Marcador de Arrays
 */
var T_ARRAY = '@';
var T_ARRAY_OBJECT = 1;
var T_ARRAY_PRIMITIVE = 0;
exports.ErrorIdRequired = 'É necessário que todos os itens de um Array de Objetos tenham identificador (id, _id, key ou _key)';
exports.ErrorArrayMixedNotAccepted = 'Não é permitido o uso de Arrays mistos (Objetos e Primitivos)';
exports.ErrorDeltaCompressHasNoDiff = 'Não é possível comprimir o objeto. Não existe diferenças. Consultar Delta.hasDiff() antes da invocação deste método';
exports.ErrorDeltaBadInitialization = 'Objeto de delta inicializado de forma errada. Consulte Delta::diff e Delta::decompress';
exports.ErrorOnlyObjectOrArrayAccepted = ' DTO aceita apenas Objetos e Arrays como entrada.';
exports.ErrorDecompressAcceptOnlyStrings = ' DTO::decompress() aceita apenas String como entrada.';
exports.ErrorArrayMultidimensionalNotAccepted = 'Não é permitido o uso de array multidimensional. Ex. key=[[],[]]';
/**
 * Data Transfer Object
 *
 * Utilitário para padronizar o transporte de dados pela rede.
 *
 * O DTO pode fazer diff e patch do objeto além de permitir comprimir os dados para transporte
 */
var DTO = /** @class */ (function () {
    function DTO(object) {
        if (object instanceof DTO) {
            object = object.unflatten();
        }
        var type = typeof object;
        if (type === 'string' || type === 'number' || type === 'boolean' || type === 'symbol' || object === null || object === undefined) {
            throw new Error(exports.ErrorOnlyObjectOrArrayAccepted);
        }
        // Evita referencia em memória
        this.object = JSON.parse(JSON.stringify(object, stringifyReplacer));
    }
    /**
     * Obtém um objeto DTO a partir de um valor comprimido
     *
     * @param compressed
     */
    DTO.decompress = function (compressed) {
        if (typeof compressed !== 'string') {
            throw new Error(exports.ErrorDecompressAcceptOnlyStrings);
        }
        // Não faz processamento agora, deixa pra fazer o processamento na invocação dos métodos
        var instance = new DTO({});
        instance.object = undefined;
        instance.compressed = compressed;
        return instance;
    };
    /**
     * Obtém um objeto DTO a partir de um valor modificado
     *
     * @param flattened
     */
    DTO.flattened = function (flattened) {
        // Não faz processamento agora, deixa pra fazer o processamento na invocação dos métodos
        var instance = new DTO({});
        instance.object = undefined;
        instance.flattened = flattened;
        return instance;
    };
    /**
    * Obtém a versão modificada do objeto original
    */
    DTO.prototype.flatten = function () {
        if (!this.flattened) {
            this.flattened = this.flattenObject(this.unflatten());
        }
        return JSON.parse(JSON.stringify(this.flattened, stringifyReplacer));
    };
    /**
     * Obtém a versão do objeto original
     */
    DTO.prototype.unflatten = function () {
        if (!this.object) {
            if (this.flattened) {
                this.object = this.unflattenObject(this.flattened);
            }
            else if (this.compressed) {
                // Decompress
                this.flattened = Compress_1.decompress(this.compressed);
                this.object = this.unflattenObject(this.flattened);
            }
        }
        return JSON.parse(JSON.stringify(this.object, stringifyReplacer));
    };
    /**
     * Gera uma versão comprimida deste objeto
     */
    DTO.prototype.compress = function () {
        if (!this.compressed) {
            this.compressed = Compress_1.compress(this.flatten());
        }
        return this.compressed;
    };
    /**
     * Permite obter a diferença entre o objeto atual e outro objeto
     *
     * @param other
     */
    DTO.prototype.diff = function (other) {
        return Delta.diff(this, other);
    };
    /**
     * Aplica um patch a esse objeto, retornando uma nova instancia do DTO
     *
     * @param delta
     */
    DTO.prototype.patch = function (delta) {
        return delta.patch(this);
    };
    /**
    * Aplica um patch (compactado) a esse objeto, retornando uma nova instancia do DTO
    *
    * @param delta
    */
    DTO.prototype.patchCompressed = function (compressedDelta) {
        return Delta.decompress(compressedDelta).patch(this);
    };
    /**
     * Remover todos os arrays de um objeto, mapeando-os por [id|_id|key|_key] em caso de objetos e INDEX para dados simples
     *
     * Valores UNDEFINED são descartados
     *
     * @param {*} obj
     */
    DTO.prototype.flattenObject = function (obj) {
        var _this = this;
        var out;
        var type = typeof obj;
        if (type === 'string' || type === 'number' || type === 'boolean' || type === 'symbol' || obj === null) {
            out = obj;
        }
        else if (Array.isArray(obj)) {
            out = {};
            var isObject_1 = undefined;
            obj.forEach(function (item, index) {
                if (Array.isArray(item)) {
                    throw new Error(exports.ErrorArrayMultidimensionalNotAccepted);
                }
                if (item === undefined) {
                    return;
                }
                var type = typeof item;
                if (type === 'string' || type === 'number' || type === 'boolean' || type === 'symbol' || item === null) {
                    // validar se existe valor mesclado
                    if (isObject_1) {
                        throw new Error(exports.ErrorArrayMixedNotAccepted);
                    }
                    isObject_1 = false;
                    // Array por índice mesmo
                    out[index] = item;
                }
                else if (type === 'object') {
                    var id_1 = item.$id || item._id || item.id || item._key || item.key;
                    if (id_1 === undefined || id_1 === null) {
                        //Essa implementação espera que todos os itens de um array possua id, _id, key ou _key.
                        throw new Error(exports.ErrorIdRequired);
                    }
                    if (isObject_1 === false) {
                        throw new Error(exports.ErrorArrayMixedNotAccepted);
                    }
                    isObject_1 = true;
                    // Se id começar com @, adiciona outra para validação no unflatten
                    var newId = (typeof (id_1) === 'string' && id_1.charAt(0) === T_ARRAY) ? (T_ARRAY + id_1) : id_1;
                    var itemFlatten = _this.flattenObject(item);
                    if (itemFlatten !== undefined) {
                        out[newId] = itemFlatten;
                    }
                }
            });
            // Tipos de array
            out[T_ARRAY] = isObject_1 ? T_ARRAY_OBJECT : T_ARRAY_PRIMITIVE;
            // Array vazio, é aceito
            // if (Object.keys(out).length === 1) {
        }
        else if (type === 'object') {
            out = {};
            for (var id in obj) {
                if (!obj.hasOwnProperty(id)) {
                    continue;
                }
                if (obj[id] === undefined) {
                    continue;
                }
                // Se id começar com @, adiciona outra para validação no unflatten
                var newId = (typeof (id) === 'string' && id.charAt(0) == T_ARRAY) ? (T_ARRAY + id) : id;
                var objIdFlatten = this.flattenObject(obj[id]);
                if (objIdFlatten !== undefined) {
                    out[newId] = objIdFlatten;
                }
            }
        }
        return out;
    };
    /**
     * Faz a decodificação do objeto anterior
     *
     * @param {*} obj
     */
    DTO.prototype.unflattenObject = function (obj) {
        var out;
        if (Array.isArray(obj)) {
            // Não aceita
            return undefined;
        }
        var type = typeof obj;
        if (type === 'string' || type === 'number' || type === 'boolean' || type === 'symbol' || obj === null) {
            // primitivo
            out = obj;
        }
        else if (type === 'object') {
            if (T_ARRAY in obj) {
                // Reverte array
                out = [];
                if (obj[T_ARRAY] === T_ARRAY_OBJECT) {
                    // Array de Objetos
                    for (var flatId in obj) {
                        if (!obj.hasOwnProperty(flatId) || flatId === T_ARRAY) {
                            continue;
                        }
                        var value = this.unflattenObject(obj[flatId]);
                        if (value === undefined) {
                            continue;
                        }
                        out.push(value);
                    }
                }
                else {
                    // Array de primitivos
                    for (var flatId in obj) {
                        if (!obj.hasOwnProperty(flatId) || flatId === T_ARRAY) {
                            continue;
                        }
                        out[Number.parseInt(flatId)] = obj[flatId];
                    }
                }
            }
            else {
                // Objeto normal
                out = {};
                for (var flatId in obj) {
                    if (!obj.hasOwnProperty(flatId)) {
                        continue;
                    }
                    var value = this.unflattenObject(obj[flatId]);
                    if (value === undefined) {
                        continue;
                    }
                    // Remove marcação de array
                    var newId = (typeof (flatId) === 'string' && flatId.charAt(0) == T_ARRAY) ? flatId.substring(1) : flatId;
                    out[newId] = value;
                }
            }
        }
        return out;
    };
    return DTO;
}());
exports.default = DTO;
;
/**
 * Representação da diferença entre dois objetos DTO
 */
var Delta = /** @class */ (function () {
    function Delta() {
        this.areEquals = undefined;
    }
    /**
     * Obtém um objeto Delta a partir da comparação de dois objetos DTO
     *
     * @param compressed
     */
    Delta.diff = function (dtoA, dtoB) {
        var instance = new Delta();
        instance.deltaObj = instance.diffObject(dtoA.flatten(), dtoB.flatten());
        return instance;
    };
    /**
     * Obtém um objeto Delta a partir de um valor comprimido
     *
     * @param compressedDelta
     */
    Delta.decompress = function (compressedDelta) {
        var instance = new Delta();
        instance.compressed = compressedDelta;
        return instance;
    };
    /**
     * Informa que não existe diferença entre os dois objetos testados
     */
    Delta.prototype.hasDiff = function () {
        if (this.areEquals === undefined) {
            this.areEquals = !(Object.keys(this.delta())[0]);
        }
        return !this.areEquals;
    };
    /**
     * Obtém a versão comprimida do DIFF, usada para tranporte
     */
    Delta.prototype.compress = function () {
        if (!this.hasDiff()) {
            throw new Error(exports.ErrorDeltaCompressHasNoDiff);
        }
        if (!this.compressed) {
            if (this.deltaObj) {
                this.compressed = Compress_1.compress(this.deltaObj);
            }
            else {
                throw new Error(exports.ErrorDeltaBadInitialization);
            }
        }
        return this.compressed;
    };
    /**
     * Obtém o objeto delta desta instancia
     */
    Delta.prototype.delta = function () {
        if (!this.deltaObj) {
            if (this.compressed) {
                this.deltaObj = Compress_1.decompress(this.compressed);
            }
            else {
                throw new Error(exports.ErrorDeltaBadInitialization);
            }
        }
        return JSON.parse(JSON.stringify(this.deltaObj, stringifyReplacer));
    };
    /**
     * Aplica esse Delta no outro objeto
     *
     * @param old
     */
    Delta.prototype.patch = function (old) {
        var flattened = old.flatten();
        this.patchObject(flattened, this.delta());
        return DTO.flattened(flattened);
    };
    /**
     * Obtém a diferença entre dois objetos (Baseado no dffptch com modificações).
     *
     * @link https://github.com/paldepind/dffptch
     *
     * @param a
     * @param b
     */
    Delta.prototype.diffObject = function (a, b) {
        var aKeys = Object.keys(a).sort();
        var bKeys = Object.keys(b).sort();
        // Adições
        var adds = {};
        // Modificações
        var mods = {};
        // Alterações recursivas
        var recs = {};
        // Remoções
        var dels = {};
        var aI = 0;
        var bI = 0;
        var delta = {};
        // Continuar o loop, desde que não cheguemos ao final das duas listas de chaves
        while (aKeys[aI] || bKeys[bI]) {
            var aKey = aKeys[aI];
            var shortAKey = String.fromCharCode(aI + 48);
            var bKey = bKeys[bI];
            var aVal = a[aKey];
            var bVal = b[bKey];
            if (aKey == bKey) {
                /**
                 * Nós estamos olhando para duas chaves iguais, isto é uma mudança - possivelmente para um objeto ou matriz
                 */
                if (Object(aVal) === aVal && Object(bVal) === bVal) {
                    // Encontre mudanças no objeto recursivamente
                    var rDelta = this.diffObject(aVal, bVal);
                    // Adicionar delta recursivo se tiver modificações
                    if (Object.keys(rDelta)[0]) {
                        recs[shortAKey] = rDelta;
                    }
                }
                else if (aVal !== bVal) {
                    mods[shortAKey] = bVal;
                }
                aI++;
                bI++;
            }
            else if (aKey > bKey || !aKey) {
                // aKey está à frente, isso significa que as chaves foram adicionadas a b
                adds[bKey] = bVal;
                bI++;
            }
            else {
                // bKey é maior, as chaves foram eliminadas
                dels[shortAKey] = 1;
                aI++;
            }
        }
        // Adicionamos somente os tipos de mudança ao delta se eles contiverem alterações
        if (Object.keys(dels)[0]) {
            delta.d = dels;
        }
        if (Object.keys(adds)[0]) {
            delta.a = adds;
        }
        if (Object.keys(mods)[0]) {
            delta.m = mods;
        }
        if (Object.keys(recs)[0]) {
            delta.r = recs;
        }
        return delta;
    };
    /**
     * Aplica o delta em um objeto (Baseado no dffptch com modificações)
     *
     * @link https://github.com/paldepind/dffptch
     * @param flattened
     * @param delta
     */
    Delta.prototype.patchObject = function (flattened, delta) {
        var operation;
        var key;
        var val;
        var longKey;
        var objKeys = Object.keys(flattened).sort();
        for (operation in delta) {
            // Operation is either 'a', 'm', 'd' or 'r'
            for (key in delta[operation]) {
                val = delta[operation][key];
                longKey = objKeys[key.charCodeAt() - 48];
                operation == 'a'
                    // addition
                    ? flattened[key] = val
                    : operation == 'm'
                        // modification
                        ? flattened[longKey] = val
                        : operation == 'd'
                            // deletion
                            ? delete flattened[longKey]
                            // recuse
                            : this.patchObject(flattened[longKey], val);
            }
        }
    };
    return Delta;
}());
exports.Delta = Delta;
/**
 * Remove os valores undefined dos arrays
 *
 * @param name
 * @param val
 */
function stringifyReplacer(name, val) {
    if (Array.isArray(val)) {
        return val.filter(function (item) { return item !== undefined; });
    }
    return val;
}
;
//# sourceMappingURL=DTO.js.map