"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
/**
 * Gerenciador de Midlewares genérico
 */
var MidlewareManager = /** @class */ (function () {
    function MidlewareManager() {
        this.middlewares = {
            '*': []
        };
    }
    /**
     * Permite adicionar Midlewares
     *
     * @param middlewares Um ou mais midlewares
     * @param name O Wildcard "*" é usado quando se deseja adicionar um midleware para todas as actions (basta não informar a action)
     */
    MidlewareManager.prototype.add = function (middlewares, name) {
        var _this = this;
        if (name === null || name === undefined) {
            name = '*';
        }
        if (!this.middlewares[name]) {
            this.middlewares[name] = [];
        }
        if (!Array.isArray(middlewares)) {
            middlewares = [middlewares];
        }
        middlewares.forEach(function (middleware) {
            if (middleware === null || middleware === undefined || typeof middleware !== 'function') {
                return;
            }
            _this.middlewares[name].push(middleware);
        });
    };
    /**
     * Permite remover um Midleware adicionado anteriormente
     *
     * @param name
     * @param middleware
     */
    MidlewareManager.prototype.remove = function (middleware, name) {
        if (name == null || name == undefined) {
            name = '*';
        }
        if (!this.middlewares[name]) {
            return;
        }
        if (name === '*') {
            for (var attr in this.middlewares) {
                if (!this.middlewares.hasOwnProperty(attr)) {
                    continue;
                }
                var idx = this.middlewares[attr].indexOf(middleware);
                if (idx >= 0) {
                    this.middlewares[attr].splice(idx, 1);
                }
            }
        }
        else {
            var idx = this.middlewares[name].indexOf(middleware);
            if (idx >= 0) {
                this.middlewares[name].splice(idx, 1);
            }
        }
    };
    /**
     * Executa os MidleWare informados
     *
     * Se o Midleware não acionar o next, o Promise nunca é resolvido (portanto, ignorado)
     *
     * @param event
     * @param context
     */
    MidlewareManager.prototype.exec = function (names, context) {
        var _this = this;
        var middlewares = (this.middlewares['*'] || []);
        if (!Array.isArray(names)) {
            names = [names];
        }
        names.forEach(function (name) {
            if (name === null || name === undefined || typeof name !== 'string') {
                return;
            }
            middlewares = middlewares.concat(_this.middlewares[name] || []);
        });
        return new Promise(function (resolve, reject) {
            // last called middleware #
            var index = -1;
            dispatch(0);
            function dispatch(actual) {
                if (actual <= index) {
                    // next() called multiple times
                    return;
                }
                index = actual;
                if (actual === middlewares.length) {
                    return resolve();
                }
                try {
                    var next = dispatch.bind(null, actual + 1);
                    middlewares[actual](context, next);
                }
                catch (err) {
                    reject(err);
                }
            }
        });
    };
    return MidlewareManager;
}());
exports.default = MidlewareManager;
//# sourceMappingURL=MidlewareManager.js.map