"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
/**
 * Classe utilitária para facilitar a observação de resolução de promises dentro de um componente
 *
 * Ex.:
 * observerLogin: PromiseObserver = new PromiseObserver(this);
 * ...
 * this.observerLogin.observe<Usuario>(LocalStorage.get('usuario')) // O estado foi atualizado
 *     .then((usuario) => {
 *        if (usuario && usuario._id) {
 *           SessaoUsuario.start(usuario as Usuario);
 *           // Está logado
 *           this.props.navigation.replace('Inicio');
 *        }
 *     });
 * ...
 * <Button
 *      isLoading={this.observerLogin.isLoding} // Bloqueio de campo até que o promise seja resolvido
 *      reverse={true}
 *      title="Entrar"
 *      onPress={this.efetuarLogin}
 *      disabled={this.state.usuario === ''}
 *      style={{ marginTop: 10 }}
 *   />
 */
var ReactIsLoading = /** @class */ (function () {
    function ReactIsLoading(component) {
        var _this = this;
        /**
         * Informa que é o primeiro carregamento observado por essa instancia
         */
        this.isFirstLoding = true;
        /**
         * Informa se existe promise sendo processado agora
         */
        this.isLoding = false;
        /**
         * Informa se possui algum erro
         */
        this.hasError = false;
        /**
         * Hash usado para permitir usar como ExtraData em listas, afim de forçar renderização de PureComponents
         */
        this.hash = (new Date()).getTime();
        this.componentUnmounted = false;
        this.promises = [];
        this.hashName = "__LoadingObserver_" + (new Date()).getTime();
        this.component = component;
        // Permite ouvir quando o componente não está mais montado, evitando processamento desnecessário
        // https://reactjs.org/blog/2015/12/16/ismounted-antipattern.html
        var oldcomponentWillUnmount = (component.componentWillUnmount || new Function).bind(component);
        component.componentWillUnmount = function () {
            _this.componentUnmounted = true;
            oldcomponentWillUnmount();
            // Remove referencia para o componente
            _this.component = undefined;
        };
    }
    /**
     * Quando o promise for resolvido, altera o estado do componente, permitindo uma nova renderização do mesmo
     *
     * O promise será finalizado antes de o componente informar o status (isLoading = true)
     *
     * @param promise
     */
    ReactIsLoading.prototype.observe = function (promise) {
        var _this = this;
        if (this.componentUnmounted) {
            // Componente não está montado
            return promise;
        }
        if (this.promises.find(function (p) { return p === promise; })) {
            // promise já existe
            return promise;
        }
        // Salva promise na lista
        this.promises.push(promise);
        // Já seta novo estado para o componente
        this.newState();
        // Ao finalizar o promise, remove da lista
        var removePromise = function () {
            var idx = _this.promises.indexOf(promise);
            if (idx >= 0) {
                _this.promises.splice(idx, 1);
            }
            _this.isFirstLoding = false;
        };
        // Observa a resolução do promise
        var received = false;
        // Garante setar novo status quando o interessado receber a resolução do promise
        var newStateOnReceive = function () {
            if (!received) {
                setTimeout(newStateOnReceive, 0);
            }
            else {
                // Define novo estado
                _this.newState();
            }
        };
        return promise
            .then(function (response) {
            removePromise();
            _this.error = undefined;
            _this.hasError = false;
            // Permite executar a cadeia do promise antes de definir novo status 
            return new Promise(function (accept) {
                if (_this.componentUnmounted) {
                    // Nunca resolve o promise, componente não está montado
                    return;
                }
                // Resolve promise
                accept(response);
                newStateOnReceive();
            });
        })
            .then(function (response) {
            received = true;
            return response;
        })
            .catch(function (reason) {
            removePromise();
            _this.error = reason;
            _this.hasError = true;
            received = true;
            // Define novo estado
            _this.newState();
            if (_this.componentUnmounted) {
                // Nunca resolve o promise, componente não está montado
                return;
            }
            // Estoura erro ocorrido
            throw reason;
        });
    };
    ReactIsLoading.prototype.newState = function () {
        // Se ainda possuir promises, está carregando
        this.isLoding = this.promises.length > 0;
        if (this.componentUnmounted || !this.component) {
            return;
        }
        var state = {};
        this.hash = (new Date()).getTime();
        state[this.hashName] = this.hash;
        this.component.setState(state);
    };
    return ReactIsLoading;
}());
exports.default = ReactIsLoading;
//# sourceMappingURL=ReactIsLoading.js.map